(in-package "OM-AURO")

(defun exactp (n)
  (ignore-errors
    (zerop (- (round n) n))))

(defmethod check-length ((lista list) (length integer) &optional (add_nil nil))
  (do ((lista- lista (cdr lista-))
       (n 0 (1+ n))
       (lista+ nil (if (car lista-) 
                     (cons (car lista-) lista+)
                     (cons 
                      (if add_nil nil
                        (car lista+)) lista+))))
      ((= n length) (reverse lista+))))

(defmethod check-length ((lista t) (length integer) &optional (add_nil nil))
  (check-length (list lista) length add_nil))

;;;;;;;; STRING TOOLS ;;;;;;;;

(defun rec-read-from-string (string)
  "utilissimo!"
  (labels ((fun (x)
             (multiple-value-list (read-from-string x nil))))
    (if (null (read-from-string string nil))
      nil
      (cons (car (fun string))
            (rec-read-from-string (coerce
                                   (nthcdr
                                    (cadr (fun string))
                                    (coerce string 'list))
                                   'string))))))

(om::defmethod! read-string ((string string))
  :icon 132
  :initvals '("Pippo sembra uno sballato")
  :indoc '("Una stringa bella complessa")
  :doc "trasforma una stringa in una lista di elementi"
  
  (rec-read-from-string string))

(om::defmethod! list->string ((lista list) &optional (directive "~D") (separator ", "))
  :icon 132
  :initvals '(("cipo" 1 2 "cacio" 3.14) "~D" ", ")
  :indoc '("Lista da convertire in stringa" 
           "Directive per la funzione Format" 
           "Separatore degli elementi della lista")
  :doc "Converte una lista in una stringa ... 
evita i noiosi format con mille direttive ..."
  
  (reduce #'om::string+
         (mapcar #'(lambda (x y)
                     (format nil (om::string+ directive y) x))
                 lista (append (make-list (1- (length lista)) :initial-element separator) '("")))))

;;;;;;;;;;;; BPF WORKS ;;;;;;;;;;;;

(setf *krate* 0.0025)

(defun bpf->lista (bpf)
  (mapcar #'list 
          (om::x-points bpf)
          (om::y-points bpf)))

(defun lista->bpf (lista &optional (dec 10))
  (let ((bpf (make-instance 'om::bpf
               :decimals dec)))
    (setf (om::x-points bpf) (mapcar #'car lista))
    (setf (om::y-points bpf) (mapcar #'cadr lista))
    bpf))

(defun diomajale (bpf)
  (let* ((xp 
          (mapcar #'(lambda (x)
                      (* *krate* (om::om-round (/ x *krate*))))
                  (om::x-points bpf)))
         (yp (om::y-points bpf))
         (dec (om::decimals bpf))
         (coppie (remove-duplicates 
                  (mapcar #'list xp yp)
                  :test #'(lambda (x y) (= (car x) (car y))))))
    (lista->bpf coppie dec)))

(defun bpf-resample (bpf &optional (krate *krate*))
  (let* ((bpf (diomajale bpf))
         (xp (om::x-points bpf))
         (lmaxxp (* krate (round (om::list-max xp) krate)))
         (lminxp (* krate (round (om::list-min xp) krate)))
         (dec (om::decimals bpf))
         (xpoints (om::arithm-ser lminxp lmaxxp krate))
         (ypoints (om::bpf-sample 
                   bpf 
                   lminxp 
                   lmaxxp
                   (length xpoints)
                   1 dec))
         (new-bpf
          (make-instance 'om::bpf
            :decimals dec)))
    (setf (om::x-points new-bpf)
          xpoints)
    (setf (om::y-points new-bpf)
          ypoints)
    new-bpf))

(defun 2bpf (bpf1 bpf2)
  (let* ((xp1 (om::x-points bpf1))
         (xp2 (om::x-points bpf2))
         (yp1 (om::y-points bpf1))
         (yp2 (om::y-points bpf2))
         (xyp1 (mapcar #'list xp1 yp1))
         (xyp2 (mapcar #'list xp2 yp2))
         (coppie1 
          (mapcar #'(lambda (x)
                      (let ((bueo (assoc (car x) xyp2)))
                        (if bueo
                          (append x (last bueo))
                          x)))
                  xyp1))
         ;(coppie2 
         ; (mapcar #'(lambda (x)
         ;             (let ((bueo (assoc (car x) xyp1)))
         ;               (when (not bueo)
         ;                 x)))
         ;         xyp2))
         )
    (om::sort. 
     (remove nil 
             coppie1
             ;(append coppie1 coppie2) 
             ) #'< #'car)))

(defun bpf+bpf (self num operation)
  (let* ((lista
          (if (equalp (om::om-round (om::x-points self) 9) (om::om-round (om::x-points num) 9))
            (mapcar #'list (om::x-points self) (om::y-points self) (om::y-points num))
            (2bpf (bpf-resample self) (bpf-resample num))))
         (lista2 
          (mapcar #'(lambda (x)
                      (list (car x) (reduce operation (cdr x))))
                  lista))
         (bpf
          (make-instance 'om::bpf
            :decimals (om::list-max (list (om::decimals self) (om::decimals num))))))
    (setf (om::x-points bpf)
          (mapcar #'car lista2))
    (setf (om::y-points bpf)
          (mapcar #'cadr lista2))
    bpf))

(defun bpf+num (self num operation)
  (let ((bpf
         (make-instance 'om::bpf
           :decimals (om::decimals self))))
    (setf (om::x-points bpf)
          (om::x-points self))
    (setf (om::y-points bpf)
          (mapcar #'(lambda (x)
                      (funcall operation x num))
                  (om::y-points self)))
    bpf))

(defun num+bpf (num self operation)
  (let ((bpf
         (make-instance 'om::bpf
           :decimals (om::decimals self))))
    (setf (om::x-points bpf)
          (om::x-points self))
    (setf (om::y-points bpf)
          (mapcar #'(lambda (x)
                      (funcall operation num x))
                  (om::y-points self)))
    bpf))

(om::defmethod* om::om+ ((self om::bpf) (num om::bpf))
  (bpf+bpf self num #'+))

(om::defmethod* om::om+ ((self list) (num om::bpf))
  (mapcar #'(lambda (x)
              (om::om+ x num)) self))

(om::defmethod* om::om+ ((self om::bpf) (num list))
  (mapcar #'(lambda (x)
              (om::om+ self x)) num))

(om::defmethod* om::om* ((self om::bpf) (num om::bpf))
  (bpf+bpf self num #'*))

(om::defmethod* om::om* ((self list) (num om::bpf))
  (mapcar #'(lambda (x)
              (om::om* x num)) self))

(om::defmethod* om::om* ((self om::bpf) (num list))
  (mapcar #'(lambda (x)
              (om::om* self x)) num))

(om::defmethod* om::om- ((self om::bpf) (num om::bpf))
  (bpf+bpf self num #'-))

(om::defmethod* om::om- ((self list) (num om::bpf))
  (mapcar #'(lambda (x)
              (om::om- x num)) self))

(om::defmethod* om::om- ((self om::bpf) (num list))
  (mapcar #'(lambda (x)
              (om::om- self x)) num))

(om::defmethod* om::om/ ((self om::bpf) (num om::bpf))
  (bpf+bpf self num #'/))

(om::defmethod* om::om/ ((self list) (num om::bpf))
  (mapcar #'(lambda (x)
              (om::om/ x num)) self))

(om::defmethod* om::om/ ((self om::bpf) (num list))
  (mapcar #'(lambda (x)
              (om::om/ self x)) num))

(om::defmethod* om::om+ ((self om::bpf) (num number))
  (bpf+num self num #'+))

(om::defmethod* om::om* ((self om::bpf) (num number))
  (bpf+num self num #'*))

(om::defmethod* om::om- ((self om::bpf) (num number))
  (bpf+num self num #'-))

(om::defmethod* om::om/ ((self om::bpf) (num number))
  (bpf+num self num #'/))

(om::defmethod* om::om+ ((self number) (num om::bpf))
  (bpf+num num self #'+))

(om::defmethod* om::om* ((self number) (num om::bpf))
  (bpf+num num self #'*))

(om::defmethod* om::om- ((self number) (num om::bpf))
  (num+bpf self num #'-))

(om::defmethod* om::om/ ((self number) (num om::bpf))
  (num+bpf self num #'/))

(om::defmethod* om::om^ ((self om::bpf) (num number))
  (bpf+num self num #'expt))

(om::defmethod* om::om^ ((self number) (num om::bpf))
  (num+bpf self num #'expt))


;;;;;;;;;;;;;;;;;;;;;;;;;;;; questo non funziona molto bene ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun bpf-simplify (bpf decimals)
  (let* ((lista (bpf->lista bpf))
         (lista2 
          (do* ((lista- lista (cdr lista-))
                (lista+ nil (cons (car lista-) lista+))
                (lista-flt nil
                           (cond
                            ((null lista-flt) (cons (car lista-) lista-flt))
                            ((= (om::om-round (cadar lista-) decimals) 
                                (om::om-round (om::om-mean (list 
                                                            (cadar lista+)
                                                            (cadadr lista-))) decimals))
                             lista-flt)
                            (t (cons (car lista-) lista-flt)))))
               ((= 2 (length lista-)) (append (reverse lista-flt) lista-))))
         (xp (mapcar #'car lista2))
         (yp (mapcar #'cadr lista2))
         (bpf2 (make-instance 'om::bpf
                 :decimals (om::decimals bpf))))
    (setf (om::x-points bpf2) xp)
    (setf (om::y-points bpf2) yp)
    bpf2))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(om::defmethod* om::mc->f  ((midics? om::bpf))
  (let ((bpf (make-instance 'om::bpf 
               :decimals (om::decimals midics?)))
        (xp (om::x-points midics?))
        (yp (om::mc->f (om::y-points midics?))))
    (setf (om::x-points bpf) xp)
    (setf (om::y-points bpf) yp)
    bpf))

(om::defmethod* om::f->mc ((freq om::bpf) &optional (approx 100) (ref-midic 0))
  (let ((bpf (make-instance 'om::bpf 
               :decimals (om::decimals freq)))
        (xp (om::x-points freq))
        (yp (mapcar #'om::f->mf (om::y-points freq))))
    (setf (om::x-points bpf) xp)
    (setf (om::y-points bpf) yp)
    bpf))

(om::defmethod* om::om-mean ((self om::bpf) &optional (weights 1))
  (om::om-mean (om::y-points self)))

(om::defmethod* om::om-mean ((self null) &optional (weights 1))
  nil)

(om::defmethod* om::om-mean ((self number) &optional (weights 1)) self)

(om::defmethod* om::om-mean ((self om::chord) &optional (weights 1))
  (float
   (om::om-mean (om::lmidic self))))

(om::defmethod* om::x-append ((l1? om::bpf) (l2? om::bpf) &rest lst?)
  (let* ((lista 
          (om::sort.
           (remove-duplicates (om::flat (mapcar #'bpf->lista (append (list l1? l2?) lst?)) 1)
                              :test #'(lambda (x y) (= (car x) (car y))))
           #'< #'car))
         (xp (mapcar #'car lista))
         (yp (mapcar #'cadr lista))
         (bpf (make-instance 'om::bpf 
                :decimals 10)))
    (setf (om::x-points bpf) xp)
    (setf (om::y-points bpf) yp)
    bpf))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; CHORD-SEQ -> BPF ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(setf *transition_curve* 0.0)

(om::defmethod! easy-envelope ((lista list) 
                                  (format integer)
                                  &optional 
                                  (krate (eval *krate*))
                                  (cntrl_type "ChannelVolume")
                                  (chan 1))
  :icon 277
  :initvals '((list (list 0.0 0.0)
                    (list 1.0 0.0)
                    (list 2.3 0.6 0.618)
                    (list 3.1 1.0 2.0)
                    (list 5.5 1.0)
                    (list 6.6 0.0))
              0 0.05 "ChannelVolume" 1)
  :menuins '((1 (("Modalys (Old)" 0) ("Modalys (New)"  1) ("BPF" 2) ("Midi Controller" 3)))
             (3 (("-- Tempo (bpm)" "Tempo")
                 ("-- KeyPress" "KeyPress")
                 ("-- ChanPress" "ChanPress")
                 ("-- PitchBend (-64 to 63)" "PitchBend")
                 ("-- PitchBend Fine (-8192 to 8191)" "PitchWheel")
                 ("00 Bank Select" "BankSelect")
                 ("01 Modulation Wheel" "ModulationWheel")
                 ("02 Breath Controller" "BreathController")
                 ("04 Foot Controller" "FootController")
                 ("05 Portamento Time" "PortamentoTime")
                 ("06 Data Entry MSB" "DataEntryMSB")
                 ("07 Channel Volume" "ChannelVolume")
                 ("08 Balance" "Balance")
                 ("10 Pan" "Pan")
                 ("11 Expression Controller" "ExpressionController")
                 ("12 Effect Control 1" "EffectControl1")
                 ("13 Effect Control 2" "EffectControl2")
                 ("16 General Purpose Controller 1" "GeneralPurposeController1")
                 ("17 General Purpose Controller 2" "GeneralPurposeController2")
                 ("18 General Purpose Controller 3" "GeneralPurposeController3")
                 ("19 General Purpose Controller 4" "GeneralPurposeController4")
                 ("00/32 Bank Select Fine" "BankSelectFine")
                 ("01/33 Modulation Wheel Fine" "ModulationWheelFine")
                 ("02/34 Breath Controller Fine" "BreathControllerFine")
                 ("04/36 Foot Controller Fine" "FootControllerFine")
                 ("05/37 Portamento Time Fine" "PortamentoTimeFine")
                 ("06/38 Data Entry MSB-LSB Fine" "DataEntryMSBLSB")
                 ("07/39 Channel Volume Fine" "ChannelVolumeFine")
                 ("08/40 Balance Fine" "BalanceFine")
                 ("10/42 Pan Fine" "PanFine")
                 ("11/43 Expression Controller Fine" "ExpressionControllerFine")
                 ("12/44 Effect Control 1 Fine" "EffectControl1Fine")
                 ("13/45 Effect Control 2 Fine" "EffectControl2Fine")
                 ("16/48 General Purpose Controller 1 Fine" "GeneralPurposeController1Fine")
                 ("17/49 General Purpose Controller 2 Fine" "GeneralPurposeController2Fine")
                 ("18/50 General Purpose Controller 3 Fine" "GeneralPurposeController3Fine")
                 ("19/51 General Purpose Controller 4 Fine" "GeneralPurposeController4Fine")
                 ("64 Damper Pedal on/off (Sustain)" "DamperPedal")
                 ("65 Portamento On/Off" "Portamento")
                 ("66 Sustenuto On/Off" "Sustenuto")
                 ("67 Soft Pedal On/Off" "SoftPedal")
                 ("68 Legato Footswitch" "LegatoFootswitch")
                 ("69 Hold 2 (on/off)" "Hold2")
                 ("70 Sound Controller 1" "SoundController1")
                 ("71 Sound Controller 2" "SoundController2")
                 ("72 Sound Controller 3" "SoundController3")
                 ("73 Sound Controller 4" "SoundController4")
                 ("74 Sound Controller 5" "SoundController5")
                 ("75 Sound Controller 6" "SoundController6")
                 ("76 Sound Controller 7" "SoundController7")
                 ("77 Sound Controller 8" "SoundController8")
                 ("78 Sound Controller 9" "SoundController9")
                 ("79 Sound Controller 10" "SoundController10")
                 ("84 Portamento Control" "PortamentoControl")
                 ("91 Effects 1 Depth" "Effects1Depth")
                 ("92 Effects 2 Depth" "Effects2Depth")
                 ("93 Effects 3 Depth" "Effects3Depth")
                 ("94 Effects 4 Depth" "Effects4Depth")
                 ("95 Effects 5 Depth" "Effects5Depth")
                 ("96 Data Increment" "DataIncrement")
                 ("-- Private" "Private"))))
  
  (let ((lista (mapcar #'(lambda (x)
                           (cons (* krate (om::om-round (/ (car x) krate)))
                                 (cdr x)))
                       lista)))
    (defun interp-2lst (lista1 lista2 kr)
      (if (or (= (cadr lista1)
                 (cadr lista2))
              (null (caddr lista2))
              (zerop (caddr lista2)))
        (list (om::first-n lista1 2)
              (om::first-n lista2 2))
        (let* ((times
                (om::arithm-ser (car lista1) (car lista2) kr))
               (vals (om::interpolation (cadr lista1) 
                                        (cadr lista2) 
                                        (length times) 
                                        (caddr lista2))))
          (mapcar #'list times vals))))
    (let*
      ((listona
        (mapcar #'(lambda (x y)
                    (interp-2lst x y krate))
                lista
                (cdr lista)))
       (listona2
        (cons (car listona)
              (mapcar #'cdr (cdr listona))))
       (listona3 (om::flat listona2 1))
       (listona4b (if (zerop (caar listona3)) 
                   listona3 
                   (cons (list 0.0 (cadar listona3)) listona3)))
       (listona4 (do ((lista- listona4b (cdr lista-))
                      (lista+ nil (if (null lista+)
                                      (cons (car lista-) lista+)
                                    (if (and (= (caar lista-) (caar lista+))
                                             (= (cadar lista-) (cadar lista+)))
                                        lista+
                                      (cons (car lista-) lista+)))))
                     ((endp lista-) (reverse lista+)))))
      (case format
        (0
         (om::mk-envelope-controller (mapcar #'car listona4)
                                     (list (mapcar #'cadr listona4))))
        ;(1
        ; (make-instance 'mlys::envelope
        ;   :dimension 1
        ;   :value listona4))
        (2
         (let ((bpf (make-instance 'om::bpf
                      :decimals 10)))
           (progn
             (setf 
              (om::x-points bpf) (mapcar #'car listona4))
             (setf 
              (om::y-points bpf) (mapcar #'cadr listona4))
             bpf)))
        (3
         (let* ((midicntrl 
                 (make-instance 'om::midicontrol
                   :ctrltype cntrl_type
                   
                   :chan chan
                   :y-points (mapcar #'cadr listona4)))
                (m2 
                 (progn
                   (setf (om::x-points midicntrl) (om::om-round (om::om* 1000.0 (mapcar #'car listona4))))
                   (setf (om::y-points midicntrl) (om::om-round (mapcar #'cadr listona4)))
                   midicntrl))
                (krat (om::om* 1000.0 krate))
                (resamp
                 (om::om-sample m2 krat)))
           resamp))))))

(om::defmethod! cseq2bpf ((cseq om::chord-seq) (output integer) 
                          &optional (transition_curve *transition_curve*) (min_interpolation_time 0.0))
  :initvals (list (make-instance 'om::chord-seq) 0 *transition_curve* 0.0)
  :indoc '("a musical sequence" "output mode" "curve for the transition between notes" "minimum interpolation time")   
  :doc "creates a BPF from a monodic sequence"
  :menuins '((1 (("Midicents" 0)("Freqs" 1))))
  :icon 128
  
  (let* ((notes (mapcar #'om::om-mean (om::lmidic cseq)))
         (onsets1 (mapcar #'(lambda (x)
                              (* *krate* (round (/ x *krate*))))
                          (om::om/ (om::lonset cseq) 1000.0)))
         (dx (om::x->dx onsets1))
         (durs 
          (mapcar #'(lambda (x)
                      (* *krate* (round (/ x *krate*))))
                  (mapcar #'om::om-mean (om::om/ (om::ldur cseq) 1000.0))))
         (durs2 (mapcar #'(lambda (x y)
                            (om::list-min (list (- x min_interpolation_time) y)))
                        dx durs))
         (alist1 (mapcar #'list onsets1 notes))
         (alist2 (mapcar #'list (om::om+ onsets1 durs2) notes))
         (alist (om::sort. (append alist1 alist2) #'< #'car))
         (lista (mapcar #'(lambda (x)
                            (append x (list transition_curve)))
                        alist))
         (bpf_mc (easy-envelope lista 2)))
    (if (zerop output)
      bpf_mc
      (om::mc->f bpf_mc))))

(om::defmethod! cseq2bpf ((cseq om::voice) (output integer) 
                          &optional (transition_curve *transition_curve*) (min_interpolation_time 0.0))
  (cseq2bpf (om::objfromobjs cseq (make-instance 'om::chord-seq)) output transition_curve min_interpolation_time))

(om::defmethod* om::Objfromobjs ((Self om::chord-seq) (Type om::bpf))
  (cseq2bpf Self 0 0))

(om::defmethod* om::Objfromobjs ((Self om::voice) (Type om::bpf))
  (cseq2bpf Self 0 0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; OM-SCALE ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(om::defmethod* om::om-scale ((self om::bpf) (minout number) (maxout number) &optional (minin 0) (maxin 0))
  (let ((bpf (om::clone self)))
    (setf (om::y-points bpf) (om::om-scale (om::y-points self) minout maxout minin maxin))
    bpf))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; PERTURBATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(om::defmethod* om::perturbation ((self om::bpf) (percent number))
  (let* ((bpf (bpf-resample self 0.01))
         (xp (om::x-points bpf))
         (yp (om::y-points bpf))
         (newbpf (make-instance 'om::bpf
                                :decimals 10))
         (newyp (om::perturbation yp percent)))
    (progn
      (setf (om::y-points newbpf) newyp)
      (setf (om::x-points newbpf) xp)
      newbpf)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun dir (dir)
  (ignore-errors
    (directory (om::string+ (namestring dir) "*") :directories t)))

(defun rec_dir (lista)
  (if
      (eval 
       (cons 'or
             (mapcar #'(lambda (x)
                         (file-directory-p x))
                     lista)))
      (rec_dir
       (remove nil
               (om::flat
                (mapcar #'(lambda (x)
                            (if (file-directory-p x)
                                (dir x)
                              x))
                        lista))))
    lista))
  
(defun all_aiff (dir)
  (remove-if-not #'(lambda (x)
                     (or 
                      (search ".aif" (namestring x))
                      (search ".AIF" (namestring x))
                      (search ".WAV" (namestring x))
                      (search ".wav" (namestring x))))
                 (rec_dir (list dir))))

(om::defmethod! folder->aiff_files ((folder_name t) &optional (sort_mode 0))
  :indoc '("a folder pathname")
  :icon 425
  :menuins '((1 (("name" 0)
                 ("creation date" 1)
                 ("sound length" 2))))
  
  (let* ((fname (cond ((stringp folder_name) (pathname folder_name))
                      ((pathnamep folder_name) folder_name)
                      (t (om::om-choose-directory-dialog))))
         ;(files (directory (format nil "~D*.aif*" fname)))
         ;(files (om::om-directory fname :type '("aif" "aiff" "AIF")))
         (files (all_aiff fname)))
    (case sort_mode 
      (0 files)
      (1 (om::sort. files #'<
                    #'(lambda (x)
                        (sys:file-stat-last-change (system::get-file-stat x)))))
      (2 (om::sort. files #'<
                    #'(lambda (x)
                        (sound-dur x)))))))

(mapcar #'(lambda (x)
            (eval
             `(om::defmethod! ,(read-from-string (format nil "unpack~D" x)) ((items list))
                :icon 425
                :initvals ',(list (om::arithm-ser 1 x 1))
                :numouts ,x
                :indoc '("list")
                :doc "a max-like unpack"
                (apply #'values items))))
        (om::arithm-ser 2 50 1))