(in-package "OM-AURO")

(defmethod mk_wave ((n_of_semi_cycles integer) (n_of_points integer) (curves t) &optional (multiplier 1.0))
  (om::om* multiplier
           (mapcar #'sin
                   (om::interpolation 
                    0 (* pi n_of_semi_cycles)
                    n_of_points curves))))

(defmethod mk_wave ((n_of_semi_cycles list) (n_of_points integer) (curves t) &optional (multiplier 1.0))
  (reduce #'om::om*
          (mapcar #'(lambda (x y z)
                      (mk_wave x n_of_points y z))
                  n_of_semi_cycles
                  (check-length curves (length n_of_semi_cycles))
                  (check-length multiplier (length n_of_semi_cycles)))))

(defun wave_envelope (lista curves expt_start &optional (expt_end nil) (expt_curve 0.0))
  (let* ((len (length lista))
         (ok (mk_wave 1 len curves))
         (env
          (if (null expt_end)
              (mapcar #'(lambda (x)
                          (expt x expt_start))
                      ok)
            (mapcar #'(lambda (x y)
                        (expt x y))
                    ok (om::interpolation expt_start expt_end len expt_curve)))))
    (om::om* env lista)))

(defun normalize_wave (lista)
  (om::om* lista (/ 1.0 (om::list-max (om::om-abs lista)))))

(defun perturb_wave (lista amount_start &optional (amount_end nil) (curve 0.0))
  (let* ((len (length lista))
         (amount
          (if (null amount_end)
              (make-list len :initial-element amount_start)
            (om::interpolation amount_start amount_end len curve)))
         (perturbated (mapcar #'om::perturbation 
                              lista
                              amount)))
    perturbated))

(om::defmethod! make_waveform ((n_of_points integer) (n_of_semi_cycles integer) (curves t) (multiplier t)
                               (curves4envelope t) (env_expt_start number) (env_expt_end number) (env_expt_curve t)
                               (perturbation_amount_start number) (perturbation_amount_end number) (perturbation_curve t)
                               &optional (output t) (size *def-table-size*))
                         
  :icon 425
  :indoc '("number of points of the BPF" "numero di semi periodi dei sinusoidi" "curva (o lista di curve) di ciascun sinusoide"
           "moltiplicatore di ciascun sinusoide" "curve per l'inviluppo globale" "expt di partenza per l'inviluppo globale"
           "expt di arrivo per l'inviluppo globale" "curva/e per l'expt dell'inviluppo" "amount of perturbation (0-1) in the beginning"
           "amount of perturbation (0-1) in the end" "curve 4 the perturbation" "output type (BPF or GEN07)")
  :doc "    create a waveform"
  :menuins `((11 (("BPF" nil)
                  ("GEN07" t)))
             (12 ,*size_menu*))
  :initvals (list 320 2 (list 0.6 2.0) 1.0 1.0 0.4 0.4 1.0 0.4 0.0 1.0 nil *def-table-size*)
  
  (let ((lista (normalize_wave
                (wave_envelope 
                 (perturb_wave 
                  (mk_wave n_of_semi_cycles n_of_points curves multiplier)
                  perturbation_amount_start
                  perturbation_amount_end
                  perturbation_curve)
                 curves4envelope
                 env_expt_start
                 env_expt_end
                 env_expt_curve)))
        (bpf (make-instance 'om::bpf
                            :decimals 10)))
    (progn
      (setf (om::y-points bpf)
            lista)
      (if output
          (m::table 7 bpf size)
        bpf))))


;;;;;;

(om::defmethod! split_soundfile ((snd t))
  (let* ((snd (inputsndfile snd))
         (n (get_n_channels snd))
         (name (no-extension (sound-name snd)))
         (sr (om::srate snd))
         (ss (om::ssize snd))
         (names (mapcar #'(lambda (x)
                            (om::string+ name "_ch" (format nil "~D" x)))
                        (om::arithm-ser 1 n 1)))
         (p3 (sound-dur snd))
         (opcodes (unpack (a_diskin2 snd 1.0))))
    (mp-mapcar #'(lambda (x)
                (easynth (car x) '(0) p3 
                         :name (cadr x)
                         :sample-rate sr
                         :sample-size ss))
            (mapcar #'list opcodes
                    names))))

(om::defmethod! split_soundfile ((snd list))
  (mapcar #'(lambda (x)
              (split_soundfile x))
          snd))

;;;;;;;

(om::defmethod! join_soundfiles ((lista list) &optional (name nil))
  (let* ((lista (mapcar #'inputsndfile lista))
         (name (if (stringp name) name (om::string+ (no-extension (sound-name (car lista))) "_join")))
         (sr (om::srate (car lista)))
         (ss (om::ssize (car lista)))
         (p3 (om::list-max (mapcar #'sound-dur lista)))
         (opcode (make_multi_opc (mapcar #'(lambda (x)
                                             (a_diskin2 x 1.0))
                                         lista))))
    (inputsndfile-om
     (easynth opcode '(0) p3 
              :name name
              :sample-rate sr
              :sample-size ss))))

(om::defmethod! mix_soundfiles ((lista list) &optional (amplist nil) (name nil) (normalize nil))
  (let* ((lista (mapcar #'inputsndfile lista))
         (name (if (stringp name) name (om::string+ (no-extension (sound-name (car lista))) "_mix")))
         (names (mapcar #'sound-name lista))
         (sr (om::srate (car lista)))
         (ss (om::ssize (car lista)))
         (nch (om::nchan (car lista)))
         (len (length lista))
         (dyn 
          (if amplist 
              (om::om-scale/sum amplist 0.8)
            (/ 0.8 len)))
         (p3 (mapcar #'sound-dur lista))
         (opcode (cs_easy-out (om::om* (a_diskin2 (p4 nch) 1.0) (p5)) nil))
         (csd (make-csd opcode (make-list len :initial-element 0) p3 names dyn)))
    (inputsndfile-om (csd->synth csd name sr ss nil normalize))))